#
# Copyright (c) 2015-2022, Sine Nomine Associates ("SNA")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND SNA DISCLAIMS ALL WARRANTIES WITH REGARD
# TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL SNA BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

# This module just stores some fabs-related constants. You can access them
# like fabs.const.VERSION, or in a dict like fabs.const.consts['VERSION'].
# The latter is just for convenience of looping through all available
# constants.

consts = {'PREFIX': '/usr', 'SYSCONFDIR': '/etc', 'LOCALSTATEDIR': '/var', 'LOCKDIR': '/var/lock', 'VERSION': '1.1', 'CONF_DIR': '/etc/fabs', 'STORAGE_DIR': '/var/lib/fabs'}

# Set fabs.const.VERSION to consts['VERSION'], etc
globals().update(consts)
